package org.toilelibre.libe.javawrapper;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.lang.reflect.Modifier;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.NotFoundException;

public class FunctionJavaWrapper {
    private final static String S             = "[\\s]*";
    private final static String IDENT         = "[a-zA-Z]\\w*";
    private final static String DEFAULT_VALUE = "(" + S + "\\=" + S + "\""
                                                      + "([^\"]*)" + "\")?";
    
    //Define here your function declaration pattern

    //R :
    private final static String FUNCTION      = "^" + S + "(" + IDENT + ")" + S
                                                      + "<-" + S + "function"
                                                      + S + "\\((" + "(?:" + S
                                                      + "(" + IDENT + ")"
                                                      + DEFAULT_VALUE + S + ","
                                                      + S + ")*" + "(?:" + S
                                                      + "(" + IDENT + ")"
                                                      + DEFAULT_VALUE + S
                                                      + ")?" + ")\\)" + S
                                                      + "[{]?" + S + "$";
    

    /*sh :
    private final static String FUNCTION      = "^" + S + "(" + IDENT + ")" + S
                                                      + "\\((" + "(?:" + S
                                                      + "(" + IDENT + ")"
                                                      + S + ","
                                                      + S + ")*" + "(?:" + S
                                                      + "(" + IDENT + ")"
                                                      + S
                                                      + ")?" + ")\\)" + S
                                                      + "[{]?" + S + "$";*/

    //Define here your default implemented invocation class
    private final static Class<? extends RWrapper> DEFAULT_INVOKE_IMPL = RWrapperRuntimeImpl.class;

    private CtClass [] createParamsArray (final CtMethod mSuperClass,
            final String [] args, final String [] defaultValues)
            throws NotFoundException {
        final ClassPool cp = ClassPool.getDefault ();
        if (mSuperClass != null) {
            return mSuperClass.getParameterTypes ();
        } else {
            final CtClass objectClass = cp.get (Object.class.getName ());
            final CtClass stringClass = cp.get (String.class.getName ());
            int length = ("".equals (args [0]) || args [0] == null ? 0 : args.length);
            final CtClass [] params = new CtClass [length];
            for (int i = 0 ; i < length ; i++) {
                if (defaultValues [i] != null) {
                    params [i] = stringClass;
                } else {
                    params [i] = objectClass;
                }
            }
            return params;
        }
    }

    private <T> CtClass defineClass (final Class<T> superclass,
            final String classname, final Class<? extends RWrapper> wrapperImpl) 
            throws CannotCompileException, NotFoundException {
        final ClassPool cp = ClassPool.getDefault ();
        final CtClass cc = cp.makeClass (classname);
        cc.addInterface (cp.get (superclass.getName ()));
        cc.setModifiers (Modifier.PUBLIC);
        final CtField ctField1 = new CtField (cp
                .get (RWrapper.class.getName ()), "wrapper", cc);
        cc.addField (ctField1, "new " + wrapperImpl.getName () + " ();");
        return cc;
    }

    private <T> CtMethod defineFunction (final CtClass clazz,
            final Class<T> superclass, final String functionName,
            final String [] args, final String [] defaultValues)
            throws NotFoundException, CannotCompileException {
        final ClassPool cp = ClassPool.getDefault ();

        final CtMethod mSuperClass = this.findBestCandidate (cp.get (superclass
                .getName ()), functionName, args);

        final CtClass [] params = this.createParamsArray (mSuperClass, args,
                defaultValues);

        final CtClass returnType = mSuperClass == null ? cp.get (Object.class
                .getName ()) : mSuperClass.getReturnType ();

        final CtMethod m = new CtMethod (returnType, functionName, params,
                clazz);
        clazz.addMethod (m);
        return m;
    }

    @SuppressWarnings ("unchecked")
    private <T> Class<T> endClassDefinition (final CtClass clazz,
            final String content) throws IOException, CannotCompileException,
            NotFoundException {
        final CtField ctField2 = new CtField (ClassPool.getDefault ().get (
                "java.lang.String"), "scriptContent", clazz);
        clazz.addField (ctField2, "\""
                + content.toString ().replace ("\n", "\\n").replace ("\"",
                        "\\\"") + "\"");
        final CtConstructor ctConstructor = new CtConstructor (new CtClass [0],
                clazz);
        clazz.addConstructor (ctConstructor);
        ctConstructor.setBody ("{" + RWrapperConf.class.getName ()
                + " rwc = this.wrapper.initLocalInstance ();"
                + "wrapper.eval (rwc, scriptContent);}");
        clazz.setModifiers (clazz.getModifiers () & ~Modifier.ABSTRACT);
        clazz.writeFile ("bin");
        return clazz.toClass ();
    }

    private CtMethod findBestCandidate (final CtClass superclass,
            final String functionName, final String [] args)
            throws NotFoundException {
        CtMethod candidate = null;
        int length = ("".equals (args [0]) || args [0] == null ? 0 : args.length);
        for (final CtMethod current : superclass.getMethods ()) {
            if (current.getParameterTypes ().length == length
                    && functionName.equals (current.getName ())) {
                candidate = current;
            }
        }
        return candidate;
    }

    private String findFileName (final File file) {
        String fileName = file.getName ();
        while (fileName.indexOf ('.') != -1) {
            fileName = fileName.substring (0, fileName.indexOf ('.'))
                    + fileName.substring (fileName.indexOf ('.') + 1,
                            fileName.indexOf ('.') + 2).toUpperCase ()
                    + fileName.substring (fileName.indexOf ('.') + 2);
        }
        if (fileName.length () > 1) {
            fileName = fileName.substring (0, 1).toUpperCase ()
                    + fileName.substring (1);
        } else {
            fileName = fileName.substring (0, 1).toUpperCase ();
        }
        return fileName;
    }

    public Object init (final File rFile) throws FileNotFoundException,
            IOException, WrapException {
        return this.init (rFile, Object.class);
    }

    public <T> T init (final File rFile, final Class<T> superclass)
            throws FileNotFoundException, IOException, WrapException {
        return this.init (rFile, superclass,
                FunctionJavaWrapper.DEFAULT_INVOKE_IMPL);
    }

    public <T> T init (final File rFile, final Class<T> superclass,
            final Class<? extends RWrapper> wrapperImpl)
            throws FileNotFoundException, IOException, WrapException {
        try {
            return this.parse (this.findFileName (rFile),
                    new FileReader (rFile), superclass, wrapperImpl)
                    .newInstance ();
        } catch (final InstantiationException e) {
            throw new WrapException (e);
        } catch (final IllegalAccessException e) {
            throw new WrapException (e);
        }
    }

    public Class<?> parse (final File rFile) throws FileNotFoundException,
            IOException, WrapException {
        return this.parse (rFile, Object.class);
    }

    public <T> Class<T> parse (final File rFile, final Class<T> superclass)
            throws FileNotFoundException, IOException, WrapException {
        return this.parse (this.findFileName (rFile), new FileReader (rFile),
                superclass);
    }

    public <T> Class<T> parse (final File rFile, final Class<T> superclass,
            final Class<? extends RWrapper> wrapperImpl)
            throws FileNotFoundException, IOException, WrapException {
        return this.parse (this.findFileName (rFile), new FileReader (rFile),
                superclass, wrapperImpl);
    }

    private <T> Class<T> parse (final String name, final Reader r,
            final Class<T> superclass) throws FileNotFoundException,
            IOException, WrapException {
        return this.parse (name, r, superclass,
                FunctionJavaWrapper.DEFAULT_INVOKE_IMPL);
    }

    private <T> Class<T> parse (final String name, final Reader r,
            final Class<T> superclass,
            final Class<? extends RWrapper> wrapperImpl)
            throws FileNotFoundException, IOException, WrapException {
        try {
            return this.parseWithExceptions (name, r, superclass, wrapperImpl);
        } catch (final CannotCompileException e) {
            throw new WrapException (e);
        } catch (final NotFoundException e) {
            throw new WrapException (e);
        }
    }

    private <T> Class<T> parseWithExceptions (final String name,
            final Reader r, final Class<T> superclass,
            final Class<? extends RWrapper> wrapperImpl)
            throws FileNotFoundException, IOException, CannotCompileException,
            NotFoundException {
        final BufferedReader br = new BufferedReader (r);
        String line = null;
        final StringBuffer content = new StringBuffer ();
        final String className = superclass.getPackage ().getName ()
                + ".generated." + name + "Impl";

        final CtClass clazz = this.defineClass (superclass, className, wrapperImpl);

        while ((line = br.readLine ()) != null) {
            content.append (line + "\n");
            final Pattern pFunction = Pattern.compile (FUNCTION);
            final Matcher mFunction = pFunction.matcher (line);
            if (mFunction.find ()) {
                this.recognizeFunction (mFunction, clazz, superclass);
            }
        }

        return this.endClassDefinition (clazz, content.toString ());
    }

    private <T> void recognizeFunction (final Matcher mFunction,
            final CtClass clazz, final Class<T> superclass)
            throws CannotCompileException, NotFoundException {
        final String functionName = mFunction.group (1);
        final String [] args = mFunction.group (2).split (",");
        int length = ("".equals (args [0]) || args [0] == null ? 0 : args.length);
        final String [] defaultValues = new String [length];
        String defaultValuesS = "";
        String argList = "";
        for (int i = 0 ; i < length ; i++) {
            argList += "\" + $" + (i + 1) + " + \"";
            if (i + 1 < length) {
                argList += ", ";
            }
            if (args [i].indexOf ('=') != -1) {
                defaultValues [i] = args [i].substring (args [i].indexOf ('='));
                final Matcher mDefaultValue = Pattern.compile (DEFAULT_VALUE)
                        .matcher (defaultValues [i]);
                if (mDefaultValue.find ()) {
                    defaultValues [i] = mDefaultValue.group (2);
                }
                args [i] = args [i].substring (0, args [i].indexOf ('='));
            }
        }

        final CtMethod method = this.defineFunction (clazz, superclass,
                functionName, args, defaultValues);
        for (int i = 0 ; i < length ; i++) {
            if (defaultValues [i] != null) {
                defaultValuesS += "if ( $" + (i + 1) + " == null) { $"
                        + (i + 1) + " = \"" + defaultValues [i] + "\"}\n";
            }
        }
        method.setBody ("{" + RWrapperConf.class.getName ()
                + " rwc = this.wrapper.initLocalInstance ();\n"
                + "String invoke = \"" + functionName + " (" + argList
                + ")\";\n" + defaultValuesS
                + "\n;return wrapper.eval (rwc, invoke);}");
    }
}
